@@ -1,13 +1,18 @@
-"""Utils to configure BigchainDB.
+# Copyright © 2020 Interplanetary Database Association e.V.,
+# BigchainDB and IPDB software contributors.
+# SPDX-License-Identifier: (Apache-2.0 AND CC-BY-4.0)
+# Code is Apache-2.0 and docs are CC-BY-4.0
 
-By calling `file_config`, the global configuration (stored in
-`$HOME/.bigchaindb`) will be updated with the values contained
-in the configuration file.
+"""Utils for reading and setting configuration settings.
 
-Note that there is a precedence in reading configuration values:
- - local config file;
- - environment vars;
- - default config file (contained in ``bigchaindb.__init__``).
+The value of each BigchainDB Server configuration setting is
+determined according to the following rules:
+
+* If it's set by an environment variable, then use that value
+* Otherwise, if it's set in a local config file, then use that
+  value
+* Otherwise, use the default value (contained in
+  ``bigchaindb.__init__``)
 """
 
 
@@ -15,13 +20,16 @@ import os
 import copy
 import json
 import logging
-import collections
+import collections.abc
 from functools import lru_cache
 
 from pkg_resources import iter_entry_points, ResolutionError
 
+from bigchaindb.common import exceptions
+
 import bigchaindb
-from bigchaindb.consensus import AbstractConsensusRules
+
+from bigchaindb.validation import BaseValidationRules
 
 # TODO: move this to a proper configuration file for logging
 logging.getLogger('requests').setLevel(logging.WARNING)
@@ -44,7 +52,7 @@ def map_leafs(func, mapping):
             path = []
 
         for key, val in mapping.items():
-            if isinstance(val, collections.Mapping):
+            if isinstance(val, collections.abc.Mapping):
                 _inner(val, path + [key])
             else:
                 mapping[key] = func(val, path=path+[key])
@@ -72,7 +80,7 @@ def update(d, u):
         mapping: An updated version of d (updated by u).
     """
     for k, v in u.items():
-        if isinstance(v, collections.Mapping):
+        if isinstance(v, collections.abc.Mapping):
             r = update(d.get(k, {}), v)
             d[k] = r
         else:
@@ -98,9 +106,14 @@ def file_config(filename=None):
 
     logger.debug('file_config() will try to open `{}`'.format(filename))
     with open(filename) as f:
-        config = json.load(f)
+        try:
+            config = json.load(f)
+        except ValueError as err:
+            raise exceptions.ConfigurationError(
+                'Failed to parse the JSON configuration from `{}`, {}'.format(filename, err)
+            )
 
-    logger.info('Configuration loaded from `{}`'.format(filename))
+        logger.info('Configuration loaded from `{}`'.format(filename))
 
     return config
 
@@ -120,6 +133,7 @@ def env_config(config):
 
     def load_from_env(value, path):
         var_name = CONFIG_SEP.join([CONFIG_PREFIX] + list(map(lambda s: s.upper(), path)))
+
         return os.environ.get(var_name, value)
 
     return map_leafs(load_from_env, config)
@@ -127,7 +141,8 @@ def env_config(config):
 
 def update_types(config, reference, list_sep=':'):
     """Return a new configuration where all the values types
-    are aligned with the ones in the default configuration"""
+    are aligned with the ones in the default configuration
+    """
 
     def _coerce(current, value):
         # Coerce a value to the `current` type.
@@ -183,11 +198,25 @@ def set_config(config):
     bigchaindb.config['CONFIGURED'] = True
 
 
+def update_config(config):
+    """Update bigchaindb.config with whatever is in the provided config dict,
+    and then set bigchaindb.config['CONFIGURED'] = True
+
+    Args:
+        config (dict): the config dict to read for changes
+                       to the default config
+    """
+
+    # Update the default config with whatever is in the passed config
+    update(bigchaindb.config, update_types(config, bigchaindb.config))
+    bigchaindb.config['CONFIGURED'] = True
+
+
 def write_config(config, filename=None):
     """Write the provided configuration to a specific location.
 
     Args:
-        newconfig (dict): a dictionary with the configuration to load.
+        config (dict): a dictionary with the configuration to load.
         filename (str): the name of the file that will store the new configuration. Defaults to ``None``.
             If ``None``, the HOME of the current user and the string ``.bigchaindb`` will be used.
     """
@@ -195,62 +224,85 @@ def write_config(config, filename=None):
         filename = CONFIG_DEFAULT_PATH
 
     with open(filename, 'w') as f:
-        json.dump(config, f)
+        json.dump(config, f, indent=4)
+
+
+def is_configured():
+    return bool(bigchaindb.config.get('CONFIGURED'))
 
 
 def autoconfigure(filename=None, config=None, force=False):
     """Run ``file_config`` and ``env_config`` if the module has not
-    been initialized."""
-
-    if not force and bigchaindb.config.get('CONFIGURED'):
+    been initialized.
+    """
+    if not force and is_configured():
         logger.debug('System already configured, skipping autoconfiguration')
         return
 
-    newconfig = env_config(bigchaindb.config)
+    # start with the current configuration
+    newconfig = bigchaindb.config
 
+    # update configuration from file
     try:
         newconfig = update(newconfig, file_config(filename=filename))
     except FileNotFoundError as e:
-        logger.warning('Cannot find config file `%s`.' % e.filename)
+        if filename:
+            raise
+        else:
+            logger.info('Cannot find config file `%s`.' % e.filename)
 
+    # override configuration with env variables
+    newconfig = env_config(newconfig)
     if config:
         newconfig = update(newconfig, config)
-
     set_config(newconfig)  # sets bigchaindb.config
 
 
 @lru_cache()
-def load_consensus_plugin(name=None):
-    """Find and load the chosen consensus plugin.
+def load_validation_plugin(name=None):
+    """Find and load the chosen validation plugin.
 
     Args:
         name (string): the name of the entry_point, as advertised in the
             setup.py of the providing package.
 
     Returns:
-        an uninstantiated subclass of ``bigchaindb.consensus.AbstractConsensusRules``
+        an uninstantiated subclass of ``bigchaindb.validation.AbstractValidationRules``
     """
     if not name:
-        name = bigchaindb.config.get('consensus_plugin', 'default')
+        return BaseValidationRules
 
-    # TODO: This will return the first plugin with group `bigchaindb.consensus`
+    # TODO: This will return the first plugin with group `bigchaindb.validation`
     #       and name `name` in the active WorkingSet.
     #       We should probably support Requirements specs in the config, e.g.
-    #       consensus_plugin: 'my-plugin-package==0.0.1;default'
+    #       validation_plugin: 'my-plugin-package==0.0.1;default'
     plugin = None
-    for entry_point in iter_entry_points('bigchaindb.consensus', name):
+    for entry_point in iter_entry_points('bigchaindb.validation', name):
         plugin = entry_point.load()
 
     # No matching entry_point found
     if not plugin:
         raise ResolutionError(
-            'No plugin found in group `bigchaindb.consensus` with name `{}`'.
+            'No plugin found in group `bigchaindb.validation` with name `{}`'.
             format(name))
 
     # Is this strictness desireable?
     # It will probably reduce developer headaches in the wild.
-    if not issubclass(plugin, (AbstractConsensusRules)):
-        raise TypeError("object of type '{}' does not implement `bigchaindb."
-                        "consensus.AbstractConsensusRules`".format(type(plugin)))
+    if not issubclass(plugin, (BaseValidationRules,)):
+        raise TypeError('object of type "{}" does not implement `bigchaindb.'
+                        'validation.BaseValidationRules`'.format(type(plugin)))
 
     return plugin
+
+
+def load_events_plugins(names=None):
+    plugins = []
+
+    if names is None:
+        return plugins
+
+    for name in names:
+        for entry_point in iter_entry_points('bigchaindb.events', name):
+            plugins.append((name, entry_point.load()))
+
+    return plugins
